/*---------------------------------------------------------------------------*\
  =========                 |
  \\      /  F ield         | OpenFOAM: The Open Source CFD Toolbox
   \\    /   O peration     |
    \\  /    A nd           | Copyright (C) 1991-2007 OpenCFD Ltd.
     \\/     M anipulation  |
-------------------------------------------------------------------------------
License
    This file is part of OpenFOAM.

    OpenFOAM is free software; you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by the
    Free Software Foundation; either version 2 of the License, or (at your
    option) any later version.

    OpenFOAM is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
    for more details.

    You should have received a copy of the GNU General Public License
    along with OpenFOAM; if not, write to the Free Software Foundation,
    Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA

\*---------------------------------------------------------------------------*/

#include "quadraticFitSnGradData.H"
#include "surfaceFields.H"
#include "volFields.H"
#include "SVD.H"
#include "geomFunctions.H"

// * * * * * * * * * * * * * * Static Data Members * * * * * * * * * * * * * //

namespace Foam
{
    defineTypeNameAndDebug(quadraticFitSnGradData, 0);
}

const Foam::scalar Foam::quadraticFitSnGradData::nExpands = 1;


// * * * * * * * * * * * * * * * * Constructors * * * * * * * * * * * * * * //

Foam::quadraticFitSnGradData::quadraticFitSnGradData
(
    const fvMesh& mesh, 
    const scalar cWeight
)
:
    MeshObject<fvMesh, quadraticFitSnGradData>(mesh),
    centralWeight_(cWeight),
#ifdef SPHERICAL_GEOMETRY
    dim_(2),
#else
    dim_(mesh.nGeometricD()),
#endif
    minSize_
    (
        dim_ == 1 ? 3 :
        dim_ == 2 ? 6 :
        dim_ == 3 ? 9 : 0
    ),
    stencilCells_(mesh.nInternalFaces()),
    stencilBounds_(mesh.nInternalFaces()),
    fit_(mesh.nInternalFaces())
{
    if (debug)
    {
        Info << "Contructing quadraticFitSnGradData" << endl;
    }

    // check input
    if (centralWeight_ < 1 - SMALL)
    {
        FatalErrorIn("quadraticFitSnGradData::quadraticFitSnGradData")
            << "centralWeight requested = " << centralWeight_
            << " should not be less than one"
            << exit(FatalError);
    }
    
    if (minSize_ == 0)
    {
        FatalErrorIn("quadraticFitSnGradData") 
            << " dimension must be 1,2 or 3, not" << dim_ << exit(FatalError);
    }

    // store the polynomial size for each face to write out
    surfaceScalarField snGradPolySize
    (
        IOobject
        (
            "quadraticFitSnGradPolySize",
            "constant",
            mesh,
            IOobject::NO_READ,
            IOobject::NO_WRITE
        ),
        mesh,
        dimensionedScalar("quadraticFitSnGradPolySize", dimless, scalar(0))
    );

    // find the stencil and the fit coefficients for every face in the mesh

    for(label faci = 0; faci < mesh.nInternalFaces(); faci++)
    {
        growStencil
        (
            stencilCells_[faci], stencilBounds_[faci], mesh, faci, nExpands
        );

        snGradPolySize[faci] = calcFit(faci);
    }
    snGradPolySize.write();
        
    if (debug)
    {
        Info<< "quadraticFitSnGradData::quadraticFitSnGradData() :"
            << "Finished constructing polynomialFit data"
            << endl;
    }
}


// * * * * * * * * * * * * * * * Member Functions  * * * * * * * * * * * * * //

Foam::label Foam::quadraticFitSnGradData::calcFit(const label faci)
{
    // const references to mesh data
    const volVectorField& C = mesh().C();
    const vectorField& Cf = mesh().faceCentres();
    
    labelList& stencilC = stencilCells_[faci];
    labelList& stencilB = stencilBounds_[faci];

    vector idir(1,0,0);
    vector jdir(0,1,0);
    vector kdir(0,0,1);
    findFaceDirs(idir, jdir, kdir, mesh(), faci);
    scalar scale = 0;

    scalarList wts(stencilC.size() + stencilB.size(), scalar(1));
    wts[0] = centralWeight_;
    wts[1] = centralWeight_;

    point p0 = Cf[faci];
    point p;
    scalar px, py, pz;
    px = py = pz = 0;

    // calculate the matrix of the polynomial components
    Matrix<scalar> B(stencilC.size() + stencilB.size(), minSize_,scalar(0));

    for(label ip = 0; ip < stencilC.size() + stencilB.size(); ip++)
    {
        label celli = ip < stencilC.size() ? stencilC[ip] 
                                           : stencilB[ip-stencilC.size()];

        if (ip < stencilC.size())
        {
            p = C[celli];
        }
        else
        {
            p = Cf[celli];
        }

        px = (p - p0)&idir;
        py = (p - p0)&jdir;
        pz = (p - p0)&kdir;
#ifdef SPHERICAL_GEOMETRY
        pz = mag(p) - mag(p0);
#endif 

        if (ip == 0) scale = max(max(mag(px), mag(py)), mag(pz));

        px /= scale;
        py /= scale;
        pz /= scale;

        label is = 0;
        B[ip][is++] = wts[ip];

        B[ip][is++] = wts[ip]*px;
        B[ip][is++] = wts[ip]*sqr(px);

        if (dim_ >= 2)
        {
            B[ip][is++] = wts[ip]*py;
            B[ip][is++] = wts[ip]*px*py;
            B[ip][is++] = wts[ip]*sqr(py);
        }
        if (dim_ == 3)
        {
            B[ip][is++] = wts[ip]*pz;
            B[ip][is++] = wts[ip]*px*pz;
            B[ip][is++] = wts[ip]*sqr(pz);
            //B[ip][is++] = wts[ip]*py*pz;
        }
    }

    SVD svd(B, SMALL);

    //Info << "face " << faci << " has singular values " << svd.S() << endl;

    label stencilSize = stencilC.size() + stencilB.size();

    fit_[faci].setSize(stencilSize);

    for(label i = 0; i < stencilSize; i++)
    {
        fit_[faci][i] = wts[i]*svd.VSinvUt()[1][i]/scale;
    }

    // remove the uncorrected snGradScheme coefficients
    //scalar maxFit = max(mag(fit_[faci]));
    fit_[faci][0] += mesh().deltaCoeffs()[faci];
    fit_[faci][1] -= mesh().deltaCoeffs()[faci];
        
    // remove cells from the stencil with ~= 0 fit coefficient
    /*label newFitSize = fit_[faci].size();
    label newCellsSize = stencilC.size();
    label newBoundsSize = stencilB.size();
     
    label i = 0;
    for(label in = 0; in < fit_[faci].size(); in++)
    {
        if (mag(fit_[faci][in])/maxFit > SMALL)
        {
            fit_[faci][i] = fit_[faci][in];
            if (in < stencilC.size())
            {
                stencilC[i] = stencilC[in];
            }
            else
            {
                stencilB[i-newCellsSize] = stencilB[in-stencilC.size()];
            }
            i++;
        }
        else
        {
            newFitSize--;
            if (in < stencilC.size())
            {
                newCellsSize--;
            }
            else
            {
                newBoundsSize--;
            }
        }
    }
        
    if (newFitSize <= 2)
    {
        newFitSize = 0;
        newCellsSize = 0;
        newBoundsSize = 0;
    }
    
    if (newFitSize < fit_[faci].size())
    {
        fit_[faci].setSize(newFitSize);
        stencilC.setSize(newCellsSize);
        stencilB.setSize(newBoundsSize);
    }
    //if (newFitSize != 0)
    //{
    //    Info << "Face " << faci << " weights " << fit_[faci] << endl
    //         << "       stencilC = " << stencilC << endl
    //         << "       stencilB = " << stencilB << endl;
    //}
    */
    return minSize_ - svd.nZeros();
}

bool Foam::quadraticFitSnGradData::movePoints()
{
    notImplemented("quadraticFitSnGradData::movePoints()");

    return true;
}


// ************************************************************************* //
